%{
#include <string.h>
#include <stdlib.h>
#include "rule_flows.h"
#include "hicuts.h"
#include "fl_test.tab.h"

#define YY_NO_UNPUT

int num_lines = 1, num_chars = 0;

int yywrap (void)
{
    return 1;
}

int yyerror (void)
{
    printf ("Error @ line %d.\n", num_lines);
    return 1;
}

void conv_mac(char *pcString, unsigned char *pcBuffer)
{
    char *pcCheck = NULL;
    char *pcPtr   = pcString;
    int   ii;

    for(ii = 0; ii < 6 ; ii++)
    {
        pcBuffer[ii] = strtoul(pcPtr, &pcCheck, 16);
        if(pcPtr == pcCheck || *pcCheck != ':')
        {
            break;
        }
        pcPtr = pcCheck + 1;
    }
}

%}


LETTER             [A-Za-z]
DIGIT              [0-9]
HEX                [0-9a-fA-F]
WHITE 		[ \t\n]
QUOTE		\"
BS		\\
PIPE		\|
AMP		\&
NEWLINE		\n
LT		\<
GT		\>
%%
{NEWLINE}        { ++num_lines; ++num_chars; }
{WHITE}          {++num_chars;}
\.               {return DOT; }
\{               {return OBR; }
\}               {return CBR; }
\/               {return SLASH; }
\(               {return OPBR; }
\)               {return CLBR; }
\:               {return COLON; }
\;               {return SEMIC; }
\?               {return QUESTION; }
\>               {return GT;}
\<               {return LT;}
\=               {return EQ;}
\>=              {return GTE;}
\<=              {return LTE;}
\!=              {return NEQ;}
<<EOF>>          {yyterminate (); }
tcp              {return TCP; }
udp              {return UDP; }
ip               {return IP;  }
nat              {return NAT; }
loadb            {return LOADB; }
normal           {return NORMAL; }
nochange         {return NOCHANGE; }
rel              {return RELATIVE; }
compute          {return COMPUTED; }
abs              {return ABSOLUTE; }
index            {return INDEX;}
value            {return VALUE;}
pass             {return PASS;}
crypto           {return CRYPTO;}
modify           {return MODIFY;}
exception        {return EXCEPTION;}
drop             {return DROP;}
none             {return NONE;}
eth[0-9]{1,2}    {
                 memset(yylval.ethintr,0,sizeof(yylval.ethintr));
                 strncpy(yylval.ethintr, yytext,5);
                 return ETHNAME;
               }
any            { num_chars += strlen (yytext); return ANY;}
permit         { num_chars += strlen (yytext); return PERMIT;}
deny           { num_chars += strlen (yytext); return DENY;}
{DIGIT}*       { yylval.num = atoi(yytext);return NUM;}
{HEX}{1,2}:{HEX}{1,2}:{HEX}{1,2}:{HEX}{1,2}:{HEX}{1,2}:{HEX}{1,2} {
                 conv_mac(yytext, yylval.mac_addr_ar);
                 return MAC_ADDR;
               }
"#"[^\n]*
end            { num_chars += strlen(yytext); return NOMORE;}
.              { ++num_chars; }
%%
