%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "rule_flows.h"
#include "hicuts.h"


rule_list *global_list;
int num_of_rules;
int per_rule_count = 0;
extern int mac_flag;

extern int yyerror (char *);
extern int yylex(void);
%}

%union {
  int num;
  rule_list *rlist;
  val_range vrange;
  rule *rle;
  rule *clist;  // This needs to be changed. 
  int ip_val;
  char ethintr[6];
  SOffset_t offset;
  unsigned char mac_addr_ar[6];
  SConditionCheck_t condcheck;
  SAction_t action;
  SCondition_t cond;
  SRule_t action_rule;
  SRule_t CAList[6];
};


%token <num> NUM SLASH DASH OBR CBR OPBR CLBR
%token DOT HOST ANY PERMIT DENY NOMORE TCP UDP IP
%token NAT LOADB NORMAL RELATIVE ABSOLUTE NOCHANGE COMPUTED 
%token QUESTION SEMIC COLON
%token NONE CRYPTO PASS MODIFY EXCEPTION DROP
%token INDEX VALUE
%token <ethintr> ETHNAME
%token LT GT EQ LTE GTE NEQ
%token <mac_addr_ar> MAC_ADDR
%type <rlist> thebase
%type <rlist > RuleList
%type <vrange> ip
%type <rle> Rule
%type <vrange> port
%type <vrange> protocol
%type <ip_val> ip_addr
%type <num> ftype
%type <offset> offset
%type <condcheck> cond_check
%type <num> op
%type <num> mod_type
%type <action> Action
%type <cond> cond
%type <action_rule> CASpec
%type <CAList> CAList
%%

thebase: RuleList { printf ("Completed parsing.\n"); global_list = $1;}
;

RuleList: { $$=NULL;}
        | Rule RuleList 
          {
	    $$ = (rule_list *) malloc (sizeof (rule_list));
	    memset($$, 0, sizeof(rule_list));
            $$->r_val = $1;
            $$->r_next = $2;

          }    
;
Rule: NUM ip ip port port protocol ftype OBR CAList CBR
      {
	int i;
          $$ = (struct rule_struct *) malloc (sizeof (struct rule_struct));
	  memset($$, 0, sizeof (struct rule_struct));
          $$->rule_id = $1;
	  $$->u.s.src_ip.flags  = $2.flags;
          $$->u.s.src_ip.low = $2.low; $$->u.s.src_ip.high = $2.high;
	  $$->u.s.dst_ip.flags  = $3.flags;
          $$->u.s.dst_ip.low = $3.low; $$->u.s.dst_ip.high = $3.high;
	  $$->u.s.src_prt.flags  = $4.flags;
          $$->u.s.src_prt.low = $4.low; $$->u.s.src_prt.high = $4.high;
	  $$->u.s.dst_prt.flags  = $5.flags;
          $$->u.s.dst_prt.low = $5.low; $$->u.s.dst_prt.high = $5.high;
	  $$->u.s.protocol.flags  = $6.flags;
          $$->u.s.protocol.low = $6.low; $$->u.s.protocol.high = $6.high;
	  $$->rule_type = $7;
	  num_of_rules++;
	  for (i = 0; i < per_rule_count; i++)
	    {
	      memcpy (&($$->rule_array[i]), &($9[i]),
		      sizeof (SRule_t));
	    }
	  per_rule_count = 0;
      }
;
ip: ip_addr SLASH NUM
    {
      $$.flags = 0;
      $$.low =  $1 & (0xFFFFFFFF << (32 - $3));
      $$.high = $1 | (0xFFFFFFFF >> $3);
    } 
  | ip_addr
    {
      $$.flags = 0;
      $$.low = $1; 
      $$.high = $1;
    }   
  | ANY
    {
      $$.flags = ANY_FIELD;
      $$.low = 0;
      $$.high = 0xFFFFFFFF;
      
    }
;

ip_addr: NUM DOT NUM DOT NUM DOT NUM
  {
    $$ = ($1 << 24) | ($3 << 16) | ($5 << 8) | $7;
  }

;

port: NUM
      {  
	$$.flags = 0;
	$$.low = $1;
	$$.high = $1;
      }
    | NUM DASH NUM
      {
	$$.flags = 0;	
	$$.low = $1;
	$$.high = $3;
      }
    | ANY
      {
	$$.flags = ANY_FIELD;
        $$.low = 0;
        $$.high = 65535;
      }
;
protocol: TCP
          {
	    $$.flags = 0;
	    $$.low = PROT_TCP;
	    $$.high = PROT_TCP;
          }
        | UDP
          {
	    $$.flags = 0;
	    $$.low = PROT_UDP;
	    $$.high = PROT_UDP;
	  }
        | IP
          {
	    $$.flags = 0;
	    $$.low = PROT_IP;
	    $$.high = PROT_IP;
	  }
    | ANY
      {
	$$.flags = ANY_FIELD;
        $$.low = 0;
        $$.high = 256;
      }
;

ftype: NORMAL
       {
	 $$ = RULE_TYPE_NORMAL;
       }
     | NAT
       {
	 $$ = RULE_TYPE_NAT;
       }
     | LOADB
       {
	 $$ = RULE_TYPE_LOADB;
       }
;

CAList: {}
      | CAList CASpec
        {
	  /*
	   * Copy this into the temporary array 
	   */
	  memcpy (&($$[per_rule_count]),
		  &($2),
		  sizeof (SRule_t));
	  per_rule_count++;
        }
;

CASpec: OPBR cond CLBR QUESTION Action COLON Action SEMIC 
       {
	 /*
	  * Copy the info into our structure.
	  */
	 memset(&($$), 0, sizeof($$));
	 memcpy (&($$.m_sOffset),     &($2.offset), sizeof(SOffset_t));
	 memcpy (&($$.m_sCondition),  &($2.cond), sizeof (SConditionCheck_t));
	 memcpy (&($$.m_sTrueAction), &($5), sizeof (SAction_t));
	 memcpy (&($$.m_sFalseAction),&($7), sizeof (SAction_t)); 
       }
;

cond: offset cond_check
      {
	/*
	 * Copy this into a temporary structure 
	 */
	memset(&($$), 0, sizeof($$));
	memcpy(&($$.offset), &($1), sizeof (SOffset_t));
	memcpy(&($$.cond), &($2), sizeof (SConditionCheck_t));
      }
;

offset: NOCHANGE
        {
	  memset(&($$), 0, sizeof($$));
	  $$.m_eOffsetType = OFFTYPE_NO_CHANGE;
	} 
      | RELATIVE NUM
        {
	  memset(&($$), 0, sizeof($$));
	  $$.m_eOffsetType = OFFTYPE_CO_RELATIVE;
	  $$.m_u32ValOrOffset = $2;
	}
      | COMPUTED NUM NUM NUM
        {
	  memset(&($$), 0, sizeof($$));
	  $$.m_eOffsetType = OFFTYPE_CO_COMPUTED;
	  $$.m_u32ValOrOffset = $2;
	  $$.m_u32Shift = $3;
	  $$.m_u32Mask = $4;
	}
      | ABSOLUTE NUM
        {
	  memset(&($$), 0, sizeof($$));
	  $$.m_eOffsetType = OFFTYPE_L3_ABSOLUTE;
	  $$.m_u32ValOrOffset = $2;
	}
;

Action: NONE
        {
	  memset(&($$), 0, sizeof($$));
	  $$.m_eActionType = FLOWACTION_NONE;
	}
      | NONE NUM
        {
	  memset(&($$), 0, sizeof($$));
	  $$.m_eActionType = FLOWACTION_NONE;
	  $$.m_uiNextRule  = $2;
	}
      | DROP
        {
	  memset(&($$), 0, sizeof($$));
	  $$.m_eActionType = FLOWACTION_DROP;
	}
      | PERMIT ETHNAME ip_addr MAC_ADDR {
	memset(&($$), 0, sizeof($$));
	$$.m_eActionType = FLOWACTION_PASS;
        $$.m_uAction.m_sParamPass.m_u32Addr = $3;
	$$.m_uAction.m_sParamPass.m_pOutDev = NULL;
        memcpy (&($$.m_uAction.m_sParamPass.m_au8DstMac),
                $4, sizeof (unsigned char) * 6);
	memcpy (&($$.m_uAction.m_sParamPass.m_acOutDevName),
		$2, 6);
      }
      | MODIFY mod_type offset NUM NUM NUM {
	memset(&($$), 0, sizeof($$));
	$$.m_eActionType = FLOWACTION_MODIFY;
        $$.m_uAction.m_sParamModify.m_u32Type = $2;
	$$.m_u32Flags = $2;
	memcpy(&($$.m_uAction.m_sParamModify.m_sOffset),
	       &($3), sizeof (SOffset_t));
	$$.m_uAction.m_sParamModify.m_u32Mask = $4;
	$$.m_uAction.m_sParamModify.m_u32ValOrIndex = $5;
	$$.m_uiNextRule  = $6;
      }
      | CRYPTO NUM {
	memset(&($$), 0, sizeof($$));
        $$.m_eActionType = FLOWACTION_CRYPTO;
	$$.m_uiNextRule  = $2;
      }
      | EXCEPTION {
	memset(&($$), 0, sizeof($$));
	$$.m_eActionType = FLOWACTION_EXCEPTION;
      }
      
;

cond_check: {
	     memset(&($$), 0, sizeof($$));
	     $$.m_u32Flags = RULE_NO_CONDITION;
             $$.mask = 0; $$.result = 0; $$.index = 0;}
          | op NUM NUM
            { 
	      memset(&($$), 0, sizeof($$));
	      $$.m_u32Flags= RULE_MASK_RESULT;
              $$.mask = $2; $$.result = $3; $$.index = 0;
              $$.op = $1;
	    }
          | op NUM NUM NUM
            {
	      memset(&($$), 0, sizeof($$));
	      $$.m_u32Flags = RULE_USE_FLOWVAR;
              $$.mask = $2; $$.result = $3; $$.index = $4;
	      $$.op = $1;
	    }
;

op: EQ
    {
      $$ = OP_EQ;
    }
  | LT
    {
      $$= OP_LT;
    }
  | GT
    {
      $$ = OP_GT;
    }

  | GTE
    {
      $$ = OP_GTE;
    }
  | LTE
    {
      $$ = OP_LTE;
    }
  | NEQ
    {
      $$ = OP_NEQ;
    }
;

mod_type: INDEX { 
                 $$ = MOD_TYPE_INDEX;
          }
        | VALUE
          {
	    $$= MOD_TYPE_VAL;
	  }
;

%%
