/*******************************************************************************

  Copyright(c) 2007 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

*******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "hash_utils.h"
#include "jhash.h"

/*
 * Hash table initialize
 *
 * This function allocates a hash table structure, 
 * and returns a pointer.
 */
struct hash_util_tab *
hash_initialize(unsigned int num_of_entries)
{
  struct hash_util_tab *entry;

  entry = (struct hash_util_tab *)malloc (sizeof (struct hash_util_tab));
  entry->hash_num = num_of_entries;
  entry->num_of_entries = 0;
  entry->num_of_coll = 0;
  
  /* Allocate the holding structure now
   */
  entry->table = (struct hash_entry_tab **)malloc(sizeof(struct hash_entry_tab) 
						 * num_of_entries);
  memset (entry->table, 0, sizeof(struct hash_entry_tab) * num_of_entries);
  return (entry);

} /* End of function hash_initialize */

/*
 * Hash_Insert
 *
 * This function inserts a hash entry
 * The key is an array of 32 bit values,
 */

RETVALUE
hash_insert (struct hash_util_tab *cntrl,
	     unsigned int *key, unsigned int keylen,
	     struct hash_entry_tab *entry)
{
  unsigned int hval;
  unsigned int tindx;

  hval = jhash2 ((void *)key, keylen, HU_INITVAL);
  tindx = hval % cntrl->hash_num;


  /*
   * Check whether we have a valid entry. We don't handle 
   * collisions now, if there is a collision, return error, 
   * else insert the entry.
   * An entry is empty is the entry is null.
   */
  
  if (cntrl->table[tindx] == NULL)
    {
      entry->hash_key = tindx;
      cntrl->table[tindx] = entry;
      cntrl->num_of_entries++;
      return RET_SUCC;
    }


  cntrl->num_of_coll++;
  return RET_FAIL;
}

RETVALUE
hash_delete(struct hash_util_tab *cntrl,
	    unsigned int *key, unsigned int keylen,
	    struct hash_entry_tab *entry)
{

  unsigned int hval;
  unsigned int tindx;

  hval = jhash2 ((void *)key, keylen, HU_INITVAL);
  tindx = hval % cntrl->hash_num;

  /*
   * Check whether we have a valid entry. We don't handle 
   * collisions now, if there is a collision, return error, 
   * else insert the entry.
   * An entry is empty is the entry is null.
   */
  
  if (cntrl->table[tindx] != NULL)
    {
      cntrl->table[tindx] = NULL;
      cntrl->num_of_entries--;
      return RET_SUCC;
    }
  return RET_FAIL;
}


struct hash_entry_tab *
hash_entry_alloc (unsigned int *key, unsigned int keylen,
		  void *entry)
{
  
  struct hash_entry_tab *retptr;

  retptr = (struct hash_entry_tab *) malloc (sizeof (struct hash_entry_tab));
  retptr->entry = entry;
  retptr->keylen = keylen;
  memcpy(retptr->key_str, key, keylen * sizeof (unsigned int));
  return (retptr);
}


struct hash_entry_tab *
hash_lookup (struct hash_util_tab *cntrl,
	     unsigned int *key, unsigned int keylen)

{
  unsigned int hval;
  unsigned int tindx;
  struct hash_entry_tab *entry;

  hval = jhash2 ((void *)key, keylen, HU_INITVAL);
  tindx = hval % cntrl->hash_num;


  entry = cntrl->table[tindx];
  
  if (entry != NULL)
    {
      if (!memcmp(entry->key_str, key, keylen * sizeof (unsigned int)))
	return entry;
      else
	return NULL;
    }
  return NULL;
}

void 
hash_printtable(struct hash_util_tab *cntrl)
{
  int i;

  printf ("Number of entries : %d\n", cntrl->num_of_entries);
  printf ("Number of collisions: %d\n", cntrl->num_of_coll);
  for (i = 0; i < cntrl->hash_num; i++)
    {
      if (cntrl->table[i] != NULL)
	{
	  printf ("\nEntry in %d \n", i);
	  hash_printkey (cntrl->table[i]);
	}
    }
}

void 
hash_printkey(struct hash_entry_tab *entry)
{
  int i;


  printf ("Printing hash key\n");
  printf ("\tHash Key Length: %d\n\t", entry->keylen);
  for (i = 0; i < entry->keylen; i++)
    {
      printf ("0x%08x ", entry->key_str[i]);
    }
  printf ("\n");
}


struct hash_entry_tab *
hash_iterator(struct hash_util_tab *cntrl, unsigned int *indx, unsigned int *cnt)
{
  int i;


  if (*cnt == cntrl->num_of_entries)
    return NULL;

  i = *indx;

  while ((i < cntrl->hash_num) & (cntrl->table[i] == NULL))
    i++;

  *indx = i;
  (*cnt)++;
  return (cntrl->table[i]);
}


void 
hash_deletetable(struct hash_util_tab *cntrl)
{
  free (cntrl->table);
  free (cntrl);
}
