/*******************************************************************************

  Copyright(c) 2007 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

*******************************************************************************/
#ifndef __HASHUTILS_H__
#define __HASHUTILS_H__


#define HU_INITVAL 0xCAFEBABE
#define RETVALUE int
#define RET_FAIL 1
#define RET_SUCC 2
#define MAX_KEY_LEN  256

/* 
 * Hash entry structure.
 * It has an array of 32 bit values as keys
 * and a pointer to an entry strcture. 
 */

struct hash_entry_tab
{
  unsigned int hash_key;  /* This is the 32 bit hash value */
  unsigned int key_str[MAX_KEY_LEN];
  unsigned int keylen;
  void *entry; /* This the hash table entry. */
  
};


/*
 * This is the hash control structure.
 */

struct hash_util_tab
{
  unsigned int hash_num; /* This is the number of entries in the hash table */
  struct hash_entry_tab **table;  /* This table is allocated when 
				    the table is initialized. */
  unsigned int num_of_entries;
  unsigned int num_of_coll;
};


struct hash_util_tab *hash_initialize(unsigned int num_of_entries);
struct hash_util_tab *hash_initialize(unsigned int num_of_entries);
RETVALUE hash_insert (struct hash_util_tab *cntrl, unsigned int *key, 
		      unsigned int keylen, struct hash_entry_tab *entry);
RETVALUE hash_delete(struct hash_util_tab *cntrl, unsigned int *key, 
		     unsigned int keylen, struct hash_entry_tab *entry);
struct hash_entry_tab *hash_entry_alloc (unsigned int *key, 
					 unsigned int keylen, void *entry);
struct hash_entry_tab *hash_lookup (struct hash_util_tab *cntrl,
	     unsigned int *key, unsigned int keylen);
void hash_printtable(struct hash_util_tab *cntrl);
struct hash_entry_tab *hash_iterator(struct hash_util_tab *cntrl, 
				     unsigned int *indx, unsigned int *cnt);
void hash_printkey(struct hash_entry_tab *entry);
void hash_deletetable(struct hash_util_tab *cntrl);
#endif /* __HASHUTILS_H__ */
