/*******************************************************************************

  Copyright(c) 2007 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

*******************************************************************************/
#ifndef __hicuts_h__
#define __hicuts_h__

/*
 * HiCuts header file
 *
 */

#define PROT_TCP   2
#define PROT_UDP   3
#define PROT_IP    4

#define MAX_RULES_PER_NODE 4
#define MIN_PARTITION_SIZE 2
#define MIN_DIM_SLICE      4

typedef enum {
  ROOT_NODE = 0,
  INTR_NODE,
  LEAF_NODE
} node_type;

typedef enum  {
  IP_SRC = 0,
  IP_DEST,
  SRC_PRT,
  DST_PRT,
  PROTOCOL
} dimension;

typedef struct val_range
{
#define ANY_FIELD 0x00000001
  unsigned int flags;
  unsigned int low;
  unsigned int high;
} val_range;


#define CHECK_FIELD(xxx, field, value) \
   ((xxx->u.s.field.flags & ANY_FIELD) || \
    (xxx->u.s.field.low  <= value && \
     xxx->u.s.field.high >= value))

/*
 * Rule Structure 
 *
 */
typedef struct rule_struct
{
  int rule_id;
#define RULE_TYPE_NORMAL  0x00
#define RULE_TYPE_NAT     0x01
#define RULE_TYPE_LOADB   0x02

  unsigned char rule_type;
  union {
    struct {
      val_range src_ip;
      val_range dst_ip;
      val_range src_prt;
      val_range dst_prt;
      val_range protocol;
    }s;
    val_range v[5];
  }u;
  SRule_t rule_array[6];
  unsigned long global_vars[4];
} rule;

/*
 * Rule List
 */
typedef struct rule_list
{
  rule *r_val;
  struct rule_list *r_next;
  struct rule_list *r_prev;
} rule_list;

/*
 * Decision Tree Node
 *
 * This is a root/intermediate/leaf node in a decision tree
 *
 */

struct decision_tree_node;
typedef struct decision_tree_node
{
  int n_type;
  rule_list r_set;       /* r_list is the list of rules NULL if not a leaf node */
  rule **r_list;         /* Used only in the driver (for convenience) */
  int num_of_rules;      /* Number of rules in the list. */
  dimension n_dim;       /* Node dimension */
  int   dim_slice;       /* Dimension slice interval */
  int   sub_tree_size;   /* Number of sub-trees in this node. */
  int   temp_index;      /* Convenience var for tree traversal */
  val_range min_max;     /* This is the min max values of the rules in 
			    this sub-node, not counting the ANY rules. 
			 */
  struct decision_tree_node *parent; /* Parent node */
  struct decision_tree_node **sub_tree; /* Sub tree */
} dtn_node;


/*
 * DTN Queue List.
 */
struct sQueue  /* Simple queue */
{
  void *node;
  struct sQueue *next;
  struct sQueue *prev;
};


struct sQueueCntrl
{
  struct sQueue head;
  struct sQueue tail;
  int num_elements;
};

#define SQUEUE_EMPTY(c) c->num_elements?0:1

/*
 * Rule List Table 
 *
 * hash table entry for rule list.
 */
typedef struct rule_hash
{
  rule *rule_ptr;
  rule *rule_next;
} rule_hash;

/*
 * Histogram Hash
 *
 */
typedef struct histo
{
  int entry_val;
  struct histo *next;
} histo;


#endif /* __hicuts_h__ */
