void print_rule (rule *rl)
{
    const char *flow_type[] = {"Normal", "NAT", "Load balance"};
    int         ii;

    MY_PRINT("Rule id (%p): %d\n", rl, rl->rule_id);
    MY_PRINT("Type: %s\n", flow_type[rl->rule_type]);

    if (rl->u.s.src_ip.flags & ANY_FIELD)
    {
        MY_PRINT("Src IP: any\n");
    }
    else
    {
        MY_PRINT("Src IP: 0x%08x 0x%08x 0x%8x\n", rl->u.s.src_ip.low,
                 rl->u.s.src_ip.high, rl->u.s.src_ip.flags);
    }
    if (rl->u.s.dst_ip.flags & ANY_FIELD)
    {
        MY_PRINT("Dst IP: any\n");
    }
    else
    {
        MY_PRINT("Dst IP: 0x%08x 0x%08x\n",
                 rl->u.s.dst_ip.low, rl->u.s.dst_ip.high);
    }
    if (rl->u.s.src_prt.flags & ANY_FIELD)
    {
        MY_PRINT("Src Port: any\n");
    }
    else
    {
        MY_PRINT("Src port: 0x%04x 0x%04x\n",
                 rl->u.s.src_prt.low, rl->u.s.src_prt.high);
    }
    if (rl->u.s.dst_prt.flags & ANY_FIELD)
    {
        MY_PRINT("Dst Port: any\n");
    }
    else
    {
        MY_PRINT("Dst port: 0x%04x 0x%04x\n",
                 rl->u.s.dst_prt.low, rl->u.s.dst_prt.high);
    }
    if (rl->u.s.protocol.flags & ANY_FIELD)
    {
        MY_PRINT("Protocol: any\n");
    }
    else
    {
        MY_PRINT("Protocol: 0x%02x 0x%02x\n",
                 rl->u.s.protocol.low, rl->u.s.protocol.high);
    }

    for(ii = 0; ii < 6; ii++)
    {
        const char *apcOffTypes[] =
            {"No change", "Relative", "Computed", "Absolute"};
        const char *apcOpTypes[] = {"==", "<", ">", ">=", "<=", "!="};
        const char *apcActTypes[] =
            {"None", "Drop", "Pass", "Modify", "Crypto", "Exception"};
        SAction_t  *pAction = NULL;
        SOffset_t  *pOffset = NULL;
        SRule_t    *pRule = NULL;
        uint32_t    u32Temp;
        int         ij;

        pRule = &rl->rule_array[ii];
        MY_PRINT("%d:\n", ii);

        MY_PRINT("\tOffset\n");
        MY_PRINT("\t\tType:\t\t%s\n",
                 (pRule->m_sOffset.m_eOffsetType < sizeof(apcOffTypes))?
                 apcOffTypes[pRule->m_sOffset.m_eOffsetType]:"???");
        MY_PRINT("\t\tValOrOffset:\t%d\n",
                 pRule->m_sOffset.m_u32ValOrOffset);
        MY_PRINT("\t\tShift:\t\t%x\n", pRule->m_sOffset.m_u32Shift);
        MY_PRINT("\t\tMask:\t\t%x\n", pRule->m_sOffset.m_u32Mask);

        MY_PRINT("\tCondition\n");
        MY_PRINT("\t\tOp:\t\t%s\n",
                 (pRule->m_sCondition.op < sizeof(apcOpTypes))?
                 apcOpTypes[pRule->m_sCondition.op]:"???");
        MY_PRINT("\t\tFlags:\t\t%x\n", pRule->m_sCondition.m_u32Flags);
        MY_PRINT("\t\tMask:\t\t%x\n", pRule->m_sCondition.mask);
        MY_PRINT("\t\tResult:\t\t%x\n", pRule->m_sCondition.result);
        MY_PRINT("\t\tIndex:\t\t%d\n", pRule->m_sCondition.index);


        for(ij = 0; ij < 2; ij++)
        {
            if(!ij)
            {
                MY_PRINT("\tTrue Action\n");
                pAction = &pRule->m_sTrueAction;
            }
            else
            {
                MY_PRINT("\tFalse Action\n");           
                pAction = &pRule->m_sFalseAction;
            }       MY_PRINT("\t\tType:\t\t%s\n",
                             (pAction->m_eActionType < sizeof(apcActTypes))?
                             apcActTypes[pAction->m_eActionType]:"???");
            MY_PRINT("\t\tFlags:\t\t%x\n", pAction->m_u32Flags);
            MY_PRINT("\t\tNext:\t\t%d\n", pAction->m_uiNextRule);
            switch(pAction->m_eActionType)
            {
                case FLOWACTION_PASS:
                    MY_PRINT("\t\tDevName:\t%s\n",
                             pAction->m_uAction.m_sParamPass.m_acOutDevName);
                    MY_PRINT("\t\tDevice:\t\t%p\n",
                             pAction->m_uAction.m_sParamPass.m_pOutDev);
                    u32Temp = pAction->m_uAction.m_sParamPass.m_u32Addr;
                    MY_PRINT("\t\tAddress:\t%d.%d.%d.%d\n",
                             (u32Temp >> 24) & 0xFF,
                             (u32Temp >> 16) & 0xFF,
                             (u32Temp >>  8) & 0xFF,
                             (u32Temp >>  0) & 0xFF);
                    MY_PRINT("\t\tMAC:\t\t%02x:%02x:%02x:%02x:%02x:%02x\n",
                             pAction->m_uAction.m_sParamPass.m_au8DstMac[0],
                             pAction->m_uAction.m_sParamPass.m_au8DstMac[1],
                             pAction->m_uAction.m_sParamPass.m_au8DstMac[2],
                             pAction->m_uAction.m_sParamPass.m_au8DstMac[3],
                             pAction->m_uAction.m_sParamPass.m_au8DstMac[4],
                             pAction->m_uAction.m_sParamPass.m_au8DstMac[5]);
                    break;
                case FLOWACTION_MODIFY:
                    pOffset = &pAction->m_uAction.m_sParamModify.m_sOffset;
                    MY_PRINT("\t\tOffType:\t%s\n",
                             (pOffset->m_eOffsetType <
                              sizeof(apcOffTypes))?
                             apcOffTypes[pOffset->m_eOffsetType]:"???");
                    MY_PRINT("\t\tOffValOrOffset:\t%d\n",
                             pOffset->m_u32ValOrOffset);
                    MY_PRINT("\t\tOffShift:\t%x\n",
                             pOffset->m_u32Shift);
                    MY_PRINT("\t\tOffMask:\t%x\n", pOffset->m_u32Mask);
                    MY_PRINT("\t\tModType:\t%x\n",
                             pAction->m_uAction.m_sParamModify.m_u32Type);
                    MY_PRINT("\t\tModMask:\t%x\n",
                             pAction->m_uAction.m_sParamModify.m_u32Mask);
                    MY_PRINT("\t\tModValOrIndex:\t%d\n",
                             pAction->m_uAction.m_sParamModify.m_u32ValOrIndex);
                    break;
                default:
                    break;
            }
        }
    }

    MY_PRINT("\n");
    MY_PRINT("\n");
}

void print_dtn_node_info (dtn_node *node, int lcnt)
{
    char *dim_str[5] = {"Src IP", "Dst IP", "Src Port",
                        "Dst Port", "Protocol"};
    char *nd_type[] = {"Root", "Intermediate", "Leaf"};
#ifdef WITHIN_PARSER
    rule_list *nxt = NULL;
#else
    int        ii;
#endif /* WITHIN_PARSER */

    MY_PRINT("\n");
    MY_PRINT("###########################: %d\n", lcnt);
    MY_PRINT("Node Type : %s\n", nd_type[node->n_type]);
    MY_PRINT("Cut Dimension: %s\n", dim_str[node->n_dim]);
    MY_PRINT("Min Value: 0x%8x Max Value: 0x%8x\n",
             node->min_max.low, node->min_max.high);
    MY_PRINT("Number of rules : %d\n", node->num_of_rules);
    MY_PRINT("Dim Slice : %d\n", node->dim_slice);
    MY_PRINT("Number of sub trees: %d\n", node->sub_tree_size);
    MY_PRINT("Rule List: \n");
    MY_PRINT("\n");
    MY_PRINT("\n");

#ifdef WITHIN_PARSER
    nxt = node->r_set.r_next;
    while (nxt != NULL)
    {
        print_rule (nxt->r_val);
        nxt = nxt->r_next;
    }
#else
    for(ii = 0; ii < node->num_of_rules;ii++)
    {
        print_rule(node->r_list[ii]);
    }
#endif /* WITHIN_PARSER */

    MY_PRINT("\n");
    MY_PRINT("----------------------------\n");
    MY_PRINT("\n");

}
