#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "rule_flows.h"
#include "hicuts.h"
#include "jhash.h"
#include "hash_utils.h"

dtn_node *glbl_root_dtn_node = NULL;
static u8 *gs_pu8XferBuffer = NULL;
static u32 gs_u32XferBufSize = 0;

int  PrintUsage(void);
void LinkDecisionTree(void);
void FreeXferBuffer(void);

int main(int argc, char *argv[])
{
    FILE *pfIn = NULL;
    long  lSize = 0;
    int   iRet = 0;
    u8   *pu8Buffer = NULL;

    if (argc != 2)
    {
	return PrintUsage();
    }

    if((pfIn = fopen(argv[1], "rb")))
    {
	fseek(pfIn, 0, SEEK_END);
	lSize = ftell(pfIn);
	rewind(pfIn);

	if((pu8Buffer = (u8 *)malloc(lSize)))
	{
	    if(lSize == fread(pu8Buffer, 1, lSize, pfIn))
	    {
		gs_pu8XferBuffer = pu8Buffer;
		gs_u32XferBufSize = lSize;
		LinkDecisionTree();
		FreeXferBuffer();
	    }
	    else
	    {
		printf("Could not read file!\n");
		iRet = 1;
	    }
	    free(pu8Buffer);
	}
	else
	{
	    printf("Could not allocate memory!\n");
	    iRet = 1;
	}

	fclose(pfIn);
    }
    else
    {
	printf("Could not open %s!\n", argv[1]);
	iRet = 1;
    }

    return iRet;
}

int PrintUsage()
{
    printf("Usage: read_test <filename>\n");
    return 1;
}

void FreeXferBuffer()
{
    if(gs_pu8XferBuffer)
    {
        dtn_node  *tmpdnode = NULL;
        dtn_node  *dnode = glbl_root_dtn_node;

        while(dnode)
        {
            if(!dnode->temp_index && dnode->r_list)
            {
                free(dnode->r_list);
                dnode->r_list = NULL;
            }

            if(dnode->temp_index < dnode->sub_tree_size)
            {
                tmpdnode = dnode->sub_tree[dnode->temp_index++];
                tmpdnode->temp_index = 0;
                dnode = tmpdnode;
            }
            else
            {
                dnode = dnode->parent;
            }
        }
    }

    gs_pu8XferBuffer = NULL;
    gs_u32XferBufSize = 0;
    glbl_root_dtn_node = NULL;
}

#define DEBUG_XFER
#ifdef DEBUG_XFER
#define MY_PRINT(format,__VARGS__...) printf(format, ##__VARGS__);
#include"obj_print.inc"
#endif /* DEBUG_XFER */

void MyPrint(void)
{
    dtn_node  *dnode = glbl_root_dtn_node;
    dtn_node  *tmpdnode = NULL;
    int        iNumNodes = 0;
    int        iNumRules = 0;
    int        iLevel = 0;

    printf("%s: %d bytes (@ %p)\n", __FUNCTION__, gs_u32XferBufSize,
           gs_pu8XferBuffer);

    dnode->temp_index = 0;

    while(dnode)
    {
        if(!dnode->temp_index)
        {
#ifdef DEBUG_XFER
            print_dtn_node_info(dnode, iLevel);
#endif /* DEBUG_XFER */
            iNumNodes++;
            iNumRules += dnode->num_of_rules;
        }

        if(dnode->temp_index < dnode->sub_tree_size)
        {
            iLevel++;
            tmpdnode = dnode->sub_tree[dnode->temp_index++];
            tmpdnode->temp_index = 0;
            dnode = tmpdnode;
        }
        else
        {
            dnode = dnode->parent;
            iLevel--;
        }
    }

    printf("%d nodes, %d rule indexes\n", iNumNodes, iNumRules);
}

void LinkDecisionTree()
{
    dtn_node *pnode = NULL;
    dtn_node *dnode = NULL;
    size_t    sz = 0;
    rule     *pRules = NULL;
    u32      *pRIndexes = NULL;
    u32       u32NumRules = 0;
    int       iIndex = 0;
    int       ii;

    // Make sure things make sense
    if(NULL == gs_pu8XferBuffer || gs_u32XferBufSize < (2 * sizeof(u32)))
    {
        return;
    }

    // get the number of rules
    iIndex += sizeof(u32);
    u32NumRules = *((u32 *) &gs_pu8XferBuffer[iIndex]);

    // get the start of the rules array
    iIndex += sizeof(u32);
    pRules = (rule *) &gs_pu8XferBuffer[iIndex];

    // Make sure that there's enough data to hold the rules
    iIndex += u32NumRules * sizeof(rule);
    if(iIndex >= gs_u32XferBufSize)
    {
        return;
    }

    for(;iIndex < gs_u32XferBufSize;)
    {
        // Make sure that there's at least enough data for the node
        dnode = (dtn_node *)&gs_pu8XferBuffer[iIndex];
        iIndex += sizeof(dtn_node);
        if(iIndex >= gs_u32XferBufSize)
        {
            return;
        }
        dnode->temp_index = 0;

        // Make sure that there's enough data for the rule indexes
        pRIndexes = (u32 *) &gs_pu8XferBuffer[iIndex];
        iIndex += dnode->num_of_rules * sizeof(u32);
        if(iIndex > gs_u32XferBufSize)
        {
            return;
        }

        // Allocate space for the rules and add them
        if(dnode->num_of_rules)
        {
            sz = dnode->num_of_rules * sizeof(rule *);
            if((dnode->r_list = (rule **) malloc(sz)))
            {
                memset(dnode->r_list, 0x00, sz);
                for(ii = 0;ii < dnode->num_of_rules;ii++)
                {
                    if(pRIndexes[ii] < u32NumRules)
                    {
                        dnode->r_list[ii] = &pRules[pRIndexes[ii]];
                    }
                }
            }
            else
            {
                return;
            }
        }

        // Link it to the tree
        if(NULL == glbl_root_dtn_node)
        {
            glbl_root_dtn_node = dnode;
        }

        if((dnode->parent = pnode))
        {
            pnode->sub_tree[pnode->temp_index++] = dnode;
        }

        // Allocate space for any subtrees
        if(dnode->sub_tree_size)
        {
            sz = dnode->sub_tree_size * sizeof(dtn_node *);
            if((dnode->sub_tree = (dtn_node **) malloc(sz)))
            {
                memset(dnode->sub_tree, 0x00, sz);
            }
            else
            {
                return;
            }
        }

        if(dnode->sub_tree_size)
        {
            pnode = dnode;
        }
        else if(pnode)
        {
            while(pnode && (pnode->temp_index >= pnode->sub_tree_size))
            {
                pnode = pnode->parent;
            }
        }

        dnode = NULL;
    }

    MyPrint();
}
