/*******************************************************************************

  Copyright(c) 2007 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

*******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "rule_flows.h"
#include "hicuts.h"
#include "jhash.h"
#include "hash_utils.h"

#define SPMF_N  15
#define MAX(A,B) A>B?A:B
#define MIN(A,B) A<=B?A:B

extern FILE *yyin;
extern rule_list *global_list;
extern int num_of_rules; /* This count is created by the parser. */
rule *g_pFlatRList = NULL;
unsigned int val_len[5] = {0xFFFFFFFF, 0xFFFFFFFF, 0xFFFF, 0xFFFF,0xFF};

#define INITVAL 0x5418138b
void print_rule (rule *rl);
int calc_unique (struct hash_util_tab *cntrl,  val_range *range);
int calc_num_partitions (rule_list *r, int num_rules,
			 dimension dim, val_range *range);
int rule_in_num_parts (rule *r, dimension dim,
		       int part_size, val_range *range);
dtn_node *create_node (rule_list *rlist, int num_rules);
struct sQueueCntrl *init_squeue ();
void sQueue_enqueue (struct sQueueCntrl *cntrl, void *entry);
void *sQueue_dequeue(struct sQueueCntrl *cntrl);
void *sQueue_iterator(struct sQueueCntrl *cntrl, int elem_num);
int max_of_dim(int a0, int a1, int a2, int a3, int a4);
void process_decision_tree_node (dtn_node *node);
void allocate_sub_tree (dtn_node *node, int num_of_parts, dimension dim);
void assign_rules_to_subtree(dtn_node *node);
void add_rule_to_node (dtn_node *node, rule *r);
void print_decision_tree(dtn_node *node);
void calc_minmax(val_range *range, int lcnt, unsigned int *minv, unsigned int *maxv);
void print_dtn_node_info (dtn_node *node, int lcnt);
dtn_node *glbl_root_dtn_node;
struct sQueueCntrl *glbl_node_queue;
void yyparse();

int GetIPAddress(const char *pcIP, unsigned int *puiAddress)
{
    unsigned int  uiTemp = 0;
    char         *pcTest = NULL;
    long          lTemp = 0;
    int           ii;

    for(ii = 0; ii < 4; ii++)
    {
        lTemp = strtol(pcIP, &pcTest, 10);

        if(lTemp >= 0 && lTemp <= 0xFF && pcIP != pcTest &&
           (((ii <  3) && '.'  == *pcTest) || ((ii == 3) && '\0' == *pcTest)))
        {
            pcIP = pcTest + 1;
        }
        else
        {
            return 0;
        }

        uiTemp |= lTemp << (24 - (ii * 8));
    }

    *puiAddress = uiTemp;

    return 1;
}

int GetPort(const char *pcPort, unsigned int *puiPort)
{
    char *pcTest = NULL;
    long  lTemp = 0;

    lTemp = strtol(pcPort, &pcTest, 0);

    if(lTemp >= 0 && lTemp <= 0xFFFF && pcPort != pcTest && '\0' == *pcTest)
    {
        *puiPort = lTemp;
        return 1;
    }

    return 0;
}

int GetProtocol(const char *pcProtocol, unsigned int *puiProtocol)
{
    if(!strcasecmp(pcProtocol,"tcp"))
    {
        *puiProtocol = PROT_TCP;
    }
    else if(!strcasecmp(pcProtocol,"udp"))
    {
        *puiProtocol = PROT_UDP;
    }
    else
    {
        return 0;
    }
    return 1;
}

void FindRule(unsigned int uiSrcIp, unsigned int uiDstIp,
              unsigned int uiSrcPort, unsigned int uiDstPort,
              unsigned int uiProtocol)
{
    unsigned int  iDimension;
    rule_list    *pRuleList = NULL;
    dtn_node     *node = glbl_root_dtn_node;
    rule         *pRule = NULL;
    int           iIndex;

    while(node && (node->num_of_rules || node->sub_tree_size))
    {
        if(node->num_of_rules)
        {
            // Try to find the rule
            for(pRuleList = node->r_set.r_next; pRuleList;
                pRuleList = pRuleList->r_next)
            {
                pRule = pRuleList->r_val;
                if(CHECK_FIELD(pRule, protocol, uiProtocol) &&
                   CHECK_FIELD(pRule, src_ip,   uiSrcIp)    &&
                   CHECK_FIELD(pRule, dst_ip,   uiDstIp)    &&
                   CHECK_FIELD(pRule, src_prt,  uiSrcPort)  &&
                   CHECK_FIELD(pRule, dst_prt,  uiDstPort))
                {
                    printf("Rule %d matches!\n", pRule->rule_id);
                    return;
                }
            }
        }

        if(node->sub_tree_size && node->sub_tree)
        {
            // figure out which branch to take
            switch(node->n_dim)
            {
                case IP_SRC:
                    iDimension = uiSrcIp;
                    break;
                case IP_DEST:
                    iDimension = uiDstIp;
                    break;
                case SRC_PRT:
                    iDimension = uiSrcPort;
                    break;
                case DST_PRT:
                    iDimension = uiDstPort;
                    break;
                case PROTOCOL:
                    iDimension = uiProtocol;
                    break;
                default:
                    printf("Invalid node!\n");
                    return;
            }

            if(iDimension < node->min_max.low ||
               iDimension > node->min_max.high)
            {
                printf("Value out of range!\n");
                break;
            }
            else
            {
                if((iIndex = ((iDimension - node->min_max.low) /
                              node->dim_slice)) == node->sub_tree_size)
                {
                    iIndex--;
                }

                node = node->sub_tree[iIndex];
            }
        }
        else
        {
            // No more nodes to search
            break;
        }
    }
    printf("Could not find a match!\n");
}

void DumpStructures(char *pcOutFilename)
{
    rule_list *nxt = NULL;
    dtn_node  *dnode = glbl_root_dtn_node;
    dtn_node  *tmpdnode = NULL;
    dtn_node   scratchNode;
    FILE      *pfOut = NULL;
    int        iNumNodes = 0;
    int        iNumRules = 0;
    int        iLevel = 0;
    u32        u32Temp;

    if(!(pfOut=fopen(pcOutFilename,"wb")))
    {
        printf("Could not open output file (%s)!\n",pcOutFilename);
        return;
    }

    dnode->temp_index = 0;

    while(dnode)
    {
        if(!dnode->temp_index)
        {
            // print_dtn_node_info(dnode, iLevel);
            iNumNodes++;
            iNumRules += dnode->num_of_rules;
        }

        if(dnode->temp_index < dnode->sub_tree_size)
        {
            iLevel++;
            tmpdnode = dnode->sub_tree[dnode->temp_index++];
            tmpdnode->temp_index = 0;
            dnode = tmpdnode;
        }
        else
        {
            dnode = dnode->parent;
            iLevel--;
        }
    }

    // Write out the over-all size (inclusive)
    u32Temp =
        // storage for length/size vars - over-all length and rule #
        (2 * sizeof(u32)) +
        // storage for the rules themselves
        (num_of_rules * sizeof(rule)) +
        // storage for the node structures
        (iNumNodes * sizeof(dtn_node)) +
        // storage for rule indexes
        (iNumRules * sizeof(u32));

    printf("%d nodes, %d rule indexes, %d bytes\n",
           iNumNodes, iNumRules, u32Temp);

    fwrite(&u32Temp, sizeof(u32Temp), 1, pfOut);

    // Write out the real rule count
    u32Temp = num_of_rules;
    fwrite(&u32Temp, sizeof(u32Temp), 1, pfOut);

    // Write out the actual rules
    fwrite(g_pFlatRList, sizeof(rule), num_of_rules, pfOut);

    // Write out the decision tree
    dnode = glbl_root_dtn_node;
    dnode->temp_index = 0;
    while(dnode)
    {
        if(!dnode->temp_index)
        {
            // Copy the node, default some values and write it out
            memcpy(&scratchNode, dnode, sizeof(dtn_node));
            memset(&scratchNode.r_set, 0x00, sizeof(rule_list));
            scratchNode.parent = NULL;
            scratchNode.sub_tree = NULL;
            fwrite(&scratchNode, sizeof(dtn_node), 1, pfOut);

            // Write out the rule index list
            nxt = dnode->r_set.r_next;
            while (nxt != NULL)
            {
                u32Temp =
                    (((u32)nxt->r_val) - ((u32)g_pFlatRList))/sizeof(rule);
                fwrite(&u32Temp, sizeof(u32Temp), 1, pfOut);
                nxt = nxt->r_next;
            }
        }

        if(dnode->temp_index < dnode->sub_tree_size)
        {
            tmpdnode = dnode->sub_tree[dnode->temp_index++];
            tmpdnode->temp_index = 0;
            dnode = tmpdnode;
        }
        else
        {
            dnode = dnode->parent;
        }
    }

    fclose(pfOut);
}

void MyPrint(void)
{
    dtn_node  *dnode = glbl_root_dtn_node;
    dtn_node  *tmpdnode = NULL;
    int        iNumNodes = 0;
    int        iNumRules = 0;
    int        iLevel = 0;

    dnode->temp_index = 0;

    while(dnode)
    {
        if(!dnode->temp_index)
        {
            print_dtn_node_info(dnode, iLevel);
            iNumNodes++;
            iNumRules += dnode->num_of_rules;
        }

        if(dnode->temp_index < dnode->sub_tree_size)
        {
            iLevel++;
            tmpdnode = dnode->sub_tree[dnode->temp_index++];
            tmpdnode->temp_index = 0;
            dnode = tmpdnode;
        }
        else
        {
            dnode = dnode->parent;
            iLevel--;
        }
    }

    printf("%d nodes, %d rule indexes\n", iNumNodes, iNumRules);

}

int main (int argc, char **argv)
{
  struct rule_list *nxt;
  struct sQueueCntrl *cntrl;
  dtn_node *dnode = NULL;

  num_of_rules = 0;

  if (argc != 2 && argc != 3 && argc != 7)
    {
      printf ("Usage: rl_parse inp_file [ out_file | <sip dip sp dp prot> ]\n");
      return 1;
    }


  yyin = fopen (argv[1],"r");
  if (yyin == NULL)
    {
      printf ("Unable to open file \n");
      return 1;
    }

  yyparse ();

  // "Flatten" the rules
  if((g_pFlatRList = (rule *)malloc(num_of_rules * sizeof(rule))))
  {
      rule_list *nxt = global_list;
      int        iIndex = 0;

      while (nxt != NULL)
      {
	  memcpy(&g_pFlatRList[iIndex],nxt->r_val,sizeof(rule));
	  free(nxt->r_val);
	  nxt->r_val = &g_pFlatRList[iIndex];
	  iIndex++;
	  nxt = nxt->r_next;
      }
  }
  else
  {
      printf("Could not allocate memory!\n");
      return 1;
  }

  printf ("The number of rules parsed is %d\n", num_of_rules);

  cntrl = init_squeue();
  glbl_node_queue = init_squeue ();

  if (global_list != NULL)
    {

      nxt = global_list;
      glbl_root_dtn_node = create_node (global_list, num_of_rules);
      glbl_root_dtn_node->n_type = ROOT_NODE;
      glbl_root_dtn_node->parent = NULL;
      sQueue_enqueue(glbl_node_queue, (void *)glbl_root_dtn_node);
      while ( !SQUEUE_EMPTY(glbl_node_queue))
      {
	dnode = sQueue_dequeue(glbl_node_queue);
	/*
	 * We need to process the queue only if we have enough
	 * rules to sub-divide.
	 * Else, we just need to move onto the next node;
	 */
	if (dnode->num_of_rules > MAX_RULES_PER_NODE)
	  {
	    process_decision_tree_node (dnode);
	    if (dnode->sub_tree_size && dnode->n_type != ROOT_NODE)
	      dnode->n_type = INTR_NODE;
	  }
      } /* End of while */
      if(argc == 2)
      {
	  // print_decision_tree(glbl_root_dtn_node);
	  MyPrint();
      }

      // Write out the rules
      if(argc == 3)
      {
	  DumpStructures(argv[2]);
      }
      else if(argc == 7)
      {
	  unsigned int uiSrcIp;
	  unsigned int uiSrcPort;
	  unsigned int uiDstIp;
	  unsigned int uiDstPort;
	  unsigned int uiProtocol;

	  if(GetIPAddress(argv[2], &uiSrcIp) &&
	     GetIPAddress(argv[3], &uiDstIp) &&
	     GetPort(argv[4], &uiSrcPort) &&
	     GetPort(argv[5], &uiDstPort) &&
	     GetProtocol(argv[6], &uiProtocol))
	  {
	      FindRule(uiSrcIp, uiDstIp, uiSrcPort, uiDstPort, uiProtocol);
	  }
	  else
	  {
	      printf("Invalid address specification\n");
	  }
      }
    }

  free(cntrl);
  free(glbl_node_queue);
  return 0;

}

/*
 * Function calculates a number of unique values.
 *
 */


int calc_unique (struct hash_util_tab *cntrl,
		  val_range *range)
{
  struct hash_entry_tab *ent;
  int ret_val;
  /*
   * First lookup the entry in the table.
   * If it is present, then return a 0,
   * else insert the entry, then return a 1.
   * We insert both ends of the range if low != high,
   * else, only on entry is inserted.
   */

  if (range->low == range->high) /* We have a point */
    {
      if (hash_lookup (cntrl, &range->low, 1))
	{
	  ret_val = 0;
	}
      else
	{
	  ent = hash_entry_alloc (&range->low, 1, (void *)range);
	  hash_insert(cntrl, &range->low, 1, ent);
	  ret_val = 1;
	}
    }
  else /* We have a segment */
    {
      /* If both ends of the segment are present, then we do not have a unique
       * value, else we have a unique value.
       */
      if (hash_lookup (cntrl, &range->low, 1))
	ret_val = 0;
      else
	{
	  ent = hash_entry_alloc (&range->low, 1, (void *)range);
	  hash_insert(cntrl, &range->low, 1, ent);
	  ret_val = 1;
	}
      if (hash_lookup (cntrl, &range->high, 1))
	ret_val += 0;
      else
	{
	  ent = hash_entry_alloc (&range->high, 1, (void *)range);
	  hash_insert(cntrl, &range->high, 1, ent);
	  ret_val += 1;
	}
    }

  if (ret_val)
    return 1;
  else
    return 0;


} /* End of calc_unique */


/*
 * Follows the HiCuts algorithm
 */

int
calc_num_partitions (rule_list *r, int num_rules, dimension dim, val_range *range)
{
  int done = 0;
  int partition_size;
  int sm_c;
  int num_part;
  rule_list *nxt;

  num_part = MAX(4,(int)sqrt((float)num_rules));

  partition_size = ((range->high - range->low) + 100)/ num_part;


  /*
   * Determine the number of partitions in the dimension.
   */
  sm_c = 0;

  while (!done)
    {
      partition_size = ((range->high - range->low) + 100)/ num_part;
      nxt = r;
      while (nxt != NULL)
	{
	  sm_c += rule_in_num_parts(nxt->r_val, dim, partition_size, range);
	  nxt = nxt->r_next;
	}
      if (sm_c < SPMF_N)
	{
	  num_part = num_part *2;
	  done = 0;
	}
      else
	done = 1;
    }

  return (num_part);
}

int
rule_in_num_parts (rule *r, dimension dim, int part_size, val_range *range)
{
  int lowindx, highindx;


  if ((r->u.v[dim].flags & ANY_FIELD) || (!part_size))
    return 0;

  lowindx = (r->u.v[dim].low - range->low) / part_size;
  highindx = (r->u.v[dim].high - range->low)/ part_size;

  return (highindx - lowindx + 1);

}


dtn_node *
create_node (rule_list *rlist, int num_rules)
{
  dtn_node *nde;

  nde = (dtn_node *)malloc (sizeof (dtn_node));
  memset(nde, 0, sizeof (dtn_node));
  nde->r_list = NULL;
  nde->n_type = LEAF_NODE;  /* default designation. */
  nde->r_set.r_next = rlist;
  nde->num_of_rules = num_rules;

  return (nde);
}


struct sQueueCntrl *
init_squeue ()
{
  struct sQueueCntrl *cntrl;
  cntrl = (struct sQueueCntrl *)malloc (sizeof (struct sQueueCntrl));
  memset(cntrl, 0, sizeof (struct sQueueCntrl));
  cntrl->head.next = &cntrl->tail;
  cntrl->head.prev = NULL;
  cntrl->tail.prev = &cntrl->head;
  cntrl->tail.next = NULL;
  cntrl->num_elements = 0;
  return cntrl;
}

void
sQueue_enqueue (struct sQueueCntrl *cntrl, void *entry)
{
  /*
   * Allocate a new queue structure and chain them up.
   */
  struct sQueue *newnode;

  newnode = (struct sQueue *) malloc (sizeof (struct sQueue));
  memset(newnode, 0, sizeof (struct sQueue));
  newnode->node = entry;
  newnode->prev = cntrl->tail.prev;
  cntrl->tail.prev->next = newnode;
  newnode->next = &cntrl->tail;
  cntrl->tail.prev = newnode;
  cntrl->num_elements++;
}

void *
sQueue_dequeue(struct sQueueCntrl *cntrl)
{
  void *entry;
  struct sQueue *nnode;

  if (cntrl->num_elements == 0)
    return NULL;

  nnode = cntrl->head.next;
  cntrl->head.next = nnode->next;

  nnode->next->prev = &cntrl->head;
  cntrl->num_elements--;
  entry = nnode->node;
  free (nnode);
  return (entry);
}

void *
sQueue_iterator(struct sQueueCntrl *cntrl, int elem_num)
{
  struct sQueue *s;
  int i;
  if (elem_num >= cntrl->num_elements)
    return NULL;

  s = cntrl->head.next;
  for (i = 0; i < elem_num; i++)
    s = s->next;

  return (s->node);

}


/*
 * Process decision tree node
 *
 * This function determines the cut dimension and partitions rules across
 * a set of child nodes.
 * If a node needs to be further processed, the node is put into a
 * queue for processing.
 *
 */
void
process_decision_tree_node (dtn_node *node)
{

  struct hash_util_tab *prot_hist;    /* Histo gram hash structures */
  struct hash_util_tab *src_prt_hist;
  struct hash_util_tab *dst_prt_hist;
  struct hash_util_tab *src_ip_hist;
  struct hash_util_tab *dst_ip_hist;
  struct rule_list *nxt;
  int uniq_prot = 0;
  int uniq_src_prt = 0;
  int uniq_dst_prt = 0;
  int uniq_src_ip = 0;
  int uniq_dst_ip = 0;
  int first_loop = 0;
  val_range minmax[5];


  prot_hist = hash_initialize(node->num_of_rules * 4);
  src_ip_hist = hash_initialize(node->num_of_rules * 4);
  dst_ip_hist = hash_initialize(node->num_of_rules * 4);
  src_prt_hist = hash_initialize(node->num_of_rules * 4);
  dst_prt_hist = hash_initialize(node->num_of_rules * 4);

  /*
   * For all the rules in this node, calculate the unique dimensions
   * and decide which one will be cut this node on.
   */

  if (node->r_set.r_next == NULL)
    return;

  nxt = node->r_set.r_next;
  while (nxt != NULL)
    {
      /*
       * Calculate the unique values as well as min and max on each.
       */
      calc_minmax(&nxt->r_val->u.s.src_prt, first_loop,
		  &minmax[SRC_PRT].low, &minmax[SRC_PRT].high);
      uniq_src_prt += calc_unique(src_prt_hist, &nxt->r_val->u.s.src_prt);

      calc_minmax(&nxt->r_val->u.s.dst_prt, first_loop,
		  &minmax[DST_PRT].low, &minmax[DST_PRT].high);
      uniq_dst_prt += calc_unique(dst_prt_hist, &nxt->r_val->u.s.dst_prt);

      calc_minmax(&nxt->r_val->u.s.src_ip, first_loop,
		  &minmax[IP_SRC].low, &minmax[IP_SRC].high);
      uniq_src_ip += calc_unique(src_ip_hist, &nxt->r_val->u.s.src_ip);

      calc_minmax(&nxt->r_val->u.s.dst_ip, first_loop,
		  &minmax[IP_DEST].low, &minmax[IP_DEST].high);
      uniq_dst_ip += calc_unique(dst_ip_hist, &nxt->r_val->u.s.dst_ip);

      calc_minmax(&nxt->r_val->u.s.protocol, first_loop,
		  &minmax[PROTOCOL].low, &minmax[PROTOCOL].high);
      uniq_prot += calc_unique(prot_hist, &nxt->r_val->u.s.protocol);
      nxt = nxt->r_next;
      first_loop++;
    }

  /*
   * We will pick the dimension which has the largest number of unique values.
   * The order is important and maps to the enum.
   */
  node->n_dim =  max_of_dim (uniq_src_ip, uniq_dst_ip, uniq_src_prt, uniq_dst_prt,
			     uniq_prot);

  node->min_max.low = minmax[node->n_dim].low;
  node->min_max.high = minmax[node->n_dim].high;

  /*
   * Now we need to determine the number of partitions that we need to make.
   */
  node->sub_tree_size = calc_num_partitions (node->r_set.r_next,
					     node->num_of_rules,
					     node->n_dim, &node->min_max);

  node->dim_slice = (node->min_max.high - node->min_max.low) / node->sub_tree_size;

  if (node->dim_slice > MIN_DIM_SLICE)
    {

      /*
       * Now that we have the partitions, allocate nodes,
       * and assign a rule to the partitions to which it belogns.
       * For each rule we check at the low and the high
       */
      allocate_sub_tree(node, node->sub_tree_size, node->n_dim);

      assign_rules_to_subtree(node);
    }
  else
  {
      node->sub_tree_size =
	  node->dim_slice = 0;
  }

} /* End of process decision tree node */


int
max_of_dim(int a0, int a1, int a2, int a3, int a4)
{
  if ((a0 >= a1) && (a0 >= a2) && (a0 >= a3) && (a0 >= a4))
    return 0;
  if ((a1 >= a0) && (a1 >= a2) && (a1 >= a3) && (a1 >= a4))
    return 1;
  if ((a2 >= a0) && (a2 >= a1) && (a2 >= a3) && (a2 >= a4))
    return 2;
  if ((a3 >= a0) && (a3 >= a1) && (a3 >= a2) && (a3 >= a4))
    return 3;
  if ((a4 >= a0) && (a4 >= a1) && (a4 >= a3) && (a4 >= a3))
    return 4;
  return 0;
}

void
allocate_sub_tree (dtn_node *node, int num_of_parts, dimension dim)
{
  int i;
  node->sub_tree = (dtn_node **)malloc (sizeof (dtn_node *) * num_of_parts);
  memset(node->sub_tree, 0, sizeof (dtn_node *) * num_of_parts);
  for (i = 0; i < num_of_parts;i++)
    {
      node->sub_tree[i] = (dtn_node *)malloc (sizeof (dtn_node));
      memset(node->sub_tree[i], 0, sizeof (dtn_node));
      memset(node->sub_tree[i], 0, sizeof (dtn_node));
      node->sub_tree[i]->parent = node;
      node->sub_tree[i]->n_type = LEAF_NODE;
    }
}

void
assign_rules_to_subtree(dtn_node *node)
{
  rule_list *nxt;
  int low_indx;
  int high_indx;
  int indx;
  rule_list *tmp;
  int i;

  nxt = node->r_set.r_next;
  while (nxt != NULL)
    {

      /* if the dimension has the ANY field set, then rule is in all
       * subtrees.
       */
      if (nxt->r_val->u.v[node->n_dim].flags & ANY_FIELD)
	{
	  for (indx = 0; indx < node->sub_tree_size; indx++)
	    {
	      add_rule_to_node(node->sub_tree[indx], nxt->r_val);
	    }

	}
      else
	{

	  /*
	   * Calculate the indices the rule is present based on the dimension.
	   */
	  low_indx = (nxt->r_val->u.v[node->n_dim].low - node->min_max.low)/ node->dim_slice;
	  high_indx = (nxt->r_val->u.v[node->n_dim].high - node->min_max.low) / node->dim_slice;

	  if (low_indx == node->sub_tree_size)
	    low_indx--;
	  if (high_indx == node->sub_tree_size)
	    high_indx--;
	
	  for (indx = low_indx; indx <= high_indx; indx++)
	    {
	      add_rule_to_node(node->sub_tree[indx], nxt->r_val);
	    }
	  /* Free up my current node.
	   */
	}
      tmp = nxt;
      nxt = nxt->r_next;
      free(tmp);
    }

  /*
   * We need to put the children into the queue
   * for further processing.
   */
  for (i = 0; i < node->sub_tree_size; i++)
    {
      sQueue_enqueue(glbl_node_queue, (void *)node->sub_tree[i]);
    }
  node->r_set.r_next = NULL;
  node->r_set.r_prev = NULL;
  node->num_of_rules = 0;

}


void
add_rule_to_node (dtn_node *node, rule *r)
{
  rule_list *new_ent;

  new_ent = (rule_list *)malloc (sizeof (rule_list));
  memset(new_ent, 0, sizeof (rule_list));
  new_ent->r_val = r;

  new_ent->r_next = node->r_set.r_next;
  new_ent->r_prev = &node->r_set;
  if (node->r_set.r_next != NULL)
    node->r_set.r_next->r_prev = new_ent;
  node->r_set.r_next = new_ent;
  node->num_of_rules++;
}


void
print_decision_tree_level(struct sQueueCntrl *print_queue, int iLevel)
{
  struct sQueueCntrl *new_print_queue;
  dtn_node *tmp;
  int i;

  new_print_queue = init_squeue ();
  while (!SQUEUE_EMPTY(print_queue))
    {
      tmp = (dtn_node *)sQueue_dequeue(print_queue);
      print_dtn_node_info (tmp, iLevel);
      for (i = 0; i < tmp->sub_tree_size; i++)
	{
	  sQueue_enqueue (new_print_queue, (void *)tmp->sub_tree[i]);
	}
    }

  if(!SQUEUE_EMPTY(new_print_queue))
  {
      print_decision_tree_level(new_print_queue,iLevel+1);
  }
  free(new_print_queue);
}

void
print_decision_tree(dtn_node *node)
{
  struct sQueueCntrl *print_queue;

  /*
   * We just iteratively print the tree.
   */
  print_queue = init_squeue ();
  sQueue_enqueue(print_queue, (void *)node);
  print_decision_tree_level(print_queue,0);
  free(print_queue);
}

void
calc_minmax(val_range *range, int lcnt, unsigned int *minv, unsigned int *maxv)
{
  if (range->flags & ANY_FIELD)
    return;
  if (lcnt == 0)
    {
      *minv = range->low;
      *maxv = range->high;
    }
  else
    {
      *minv = MIN(*minv, range->low);
      *maxv = MAX(*maxv, range->high);
    }
} /* End of calc_minmax */

#define MY_PRINT(format,__VARGS__...) printf(format, ##__VARGS__);
#define WITHIN_PARSER 1
#include"obj_print.inc"
