/*******************************************************************************

  Copyright(c) 2007 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

*******************************************************************************/
#ifndef _RULE_FLOWS_H_
#define _RULE_FLOWS_H_

#ifndef __KERNEL__
#include <stdint.h>
#endif /* __KERNEL__ */

typedef enum
{
    OFFTYPE_NO_CHANGE=0,
    OFFTYPE_CO_RELATIVE,
    OFFTYPE_CO_COMPUTED,
    OFFTYPE_L3_ABSOLUTE,
} EOffset_t;

typedef struct
{
    EOffset_t m_eOffsetType;
    uint32_t  m_u32ValOrOffset;
    uint32_t  m_u32Shift;
    uint32_t  m_u32Mask;
} SOffset_t;

typedef enum
{
    FLOWACTION_NONE=0,
    FLOWACTION_DROP,
    FLOWACTION_PASS,
    FLOWACTION_MODIFY,
    FLOWACTION_CRYPTO,
    FLOWACTION_EXCEPTION,
} EFlowAction_t;

typedef struct
{
  unsigned char      m_acOutDevName[6];
  void              *m_pOutDev;
  uint32_t           m_u32Addr;
#define ETH_ALEN 6
  uint8_t            m_au8DstMac[ETH_ALEN];
} SParamPass;

typedef struct
{
  SOffset_t          m_sOffset;
  uint32_t           m_u32Type;
#define MOD_TYPE_INDEX (1<<0)
#define MOD_TYPE_VAL   (1<<1)

  uint32_t           m_u32Mask;
  uint32_t           m_u32ValOrIndex;
} SParamModify;

typedef struct
{
    EFlowAction_t m_eActionType;

    uint32_t  m_u32Flags;
#define ACTION_MOD_USE_GLOBAL_VALS (1<<0)
#define ACTION_PRINTED_DEV_ERROR   (1<<1)

    union
    {
      SParamPass   m_sParamPass;
      SParamModify m_sParamModify;
    } m_uAction;
    // Jump (if applicable, based on Action type)
    unsigned int m_uiNextRule;
} SAction_t;

typedef enum
{
    OP_EQ=0,
    OP_LT,
    OP_GT,
    OP_GTE,
    OP_LTE,
    OP_NEQ,
} EOP_t;
 
typedef struct
{
  EOP_t op;
  uint32_t  m_u32Flags;
#define RULE_MASK_RESULT  (1<<0) // Check Mask& Result 
#define RULE_NO_CONDITION (1<<1) // Perform True action by default
#define RULE_USE_FLOWVAR  (1<<2)

  unsigned int mask;
  unsigned int result;
  unsigned int index;
} SConditionCheck_t;

/* This is a temp structure for parsing purposes.
 */
typedef struct 
{
  SOffset_t offset;
  SConditionCheck_t cond;
} SCondition_t;

typedef struct
{
  // Offset Action
  SOffset_t m_sOffset;
  
  SConditionCheck_t  m_sCondition;
  
  // True Action
  SAction_t m_sTrueAction;
  
  // False Action
  SAction_t m_sFalseAction;
} SRule_t;

#endif /* _RULE_FLOWS_H_ */
