/*******************************************************************************

  Copyright(c) 2007 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

*******************************************************************************/
#ifndef _FLOWS_H_
#define _FLOWS_H_

#include "../../HiCuts/rule_flows.h"

#define FLOW_MAX_CARDS                  20
#define FLOW_MAX_CARDS_START            0
#define FLOW_MAX_HASH_LEVELS            4
#define FLOW_MAX_FLOWS                  500000
#define FLOW_MAX_2NDFLOWS               (FLOW_MAX_FLOWS / 2)
#define FLOW_MAX_3RDFLOWS               (FLOW_MAX_FLOWS / 4)
#define FLOW_MAX_4THFLOWS               (FLOW_MAX_FLOWS / 8)
#define FLOW_MAX_AGE_MINUTES            10
#define FLOW_MAX_RULES_PER              6
#define FLOW_MAX_VALUES_PER             4

#define FLOW_MAX_NAT_PORTS              2000
#define FLOW_NAT_PORT_START             2048

#define FLOW_1STLVL_INITVAL             93047
#define FLOW_2NDLVL_INITVAL             59834
#define FLOW_3RDLVL_INITVAL             26415
#define FLOW_4THLVL_INITVAL             12308

#define FLOW_ALLOC_BLOCK_SIZE           64

#ifdef FLOWS_DEBUG
#define DEBUG_SAMPLES                   100
#endif /* FLOWS_DEBUG */

struct sk_buff;

#define TUPPLE_U32_LEN                  4

typedef union
{
    uint32_t m_au32Data[TUPPLE_U32_LEN];
    struct
    {
        uint32_t m_u32DstAddr;
        uint32_t m_u32SrcAddr;
        uint16_t m_u16DstPort;
        uint16_t m_u16SrcPort;
        uint8_t  m_u8Protocol;
    } m_sTuppleBits;
} UTupple_t;

typedef struct
{
    uint32_t        m_u32Flags;
#define FLOW_FLAG_NAT   (1<<0)
#define FLOW_FLAG_LOADB (1<<1)
#define FLOW_FLAG_PAIR  (1<<2)
#define FLOW_COLLISION  (1<<3)

    struct timeval m_tvLastAccess;
    SRule_t        m_aRules[FLOW_MAX_RULES_PER];
    uint32_t       m_au32Values[FLOW_MAX_VALUES_PER];
    UTupple_t      m_uTupple;
    UTupple_t      m_uPairTupple;
    struct net_device *m_pPairNetDev;
} SFlow_t;

typedef enum
{
    FLOW_Q_TYPE_EMPTY = 0,
    FLOW_Q_TYPE_NORMAL,
    FLOW_Q_TYPE_EXCEPTION,
    FLOW_Q_TYPE_DEL,
} EQEntryType_t;

typedef struct
{
    EQEntryType_t   m_eType;
    UTupple_t       m_uTupple;
    struct sk_buff *skb;
    struct net_device *m_pNetDev;
} SQEntry_t;

#define FLOW_Q_DEPTH                    1024

typedef struct
{
    int                 m_iIndex;

    struct task_struct *m_pTask;
    struct completion   m_sStartComp;
    struct completion   m_sExitComp;

    int                 m_iWIndex;
    int                 m_iRIndex;
    SQEntry_t           m_asQ[FLOW_Q_DEPTH];
#ifdef FLOWS_DEBUG
    cycles_t            m_acySamples[DEBUG_SAMPLES];
    cycles_t            m_cyMin;
    cycles_t            m_cyMax;
    int                 m_iNumSamples;
    int                 m_iSampleIndex;
#endif /* FLOWS_DEBUG */
} SThreadCtxt_t;

/* Changed to kmem_cache from kmem_cache_t for 2.6.27 -SMO */
extern struct kmem_cache *g_pActionMemCache;

typedef struct
{
    int       m_iSize;
    u32       m_u32InitVal;
    u32       m_u32Cols;
    SFlow_t **m_apFlows;
} SHashLevel_t;

#define INIT_HASH_LEVELS(x) \
{ \
    (x)[0].m_iSize      = FLOW_MAX_FLOWS; \
    (x)[0].m_u32InitVal = FLOW_1STLVL_INITVAL; \
    (x)[0].m_u32Cols    = 0; \
    (x)[0].m_apFlows    = NULL; \
 \
    (x)[1].m_iSize      = FLOW_MAX_2NDFLOWS; \
    (x)[1].m_u32InitVal = FLOW_2NDLVL_INITVAL; \
    (x)[1].m_u32Cols    = 0; \
    (x)[1].m_apFlows    = NULL; \
 \
    (x)[2].m_iSize      = FLOW_MAX_3RDFLOWS; \
    (x)[2].m_u32InitVal = FLOW_3RDLVL_INITVAL; \
    (x)[2].m_u32Cols    = 0; \
    (x)[2].m_apFlows    = NULL; \
 \
    (x)[3].m_iSize      = FLOW_MAX_4THFLOWS; \
    (x)[3].m_u32InitVal = FLOW_4THLVL_INITVAL; \
    (x)[3].m_u32Cols    = 0; \
    (x)[3].m_apFlows    = NULL; \
}

void DoIt(struct sk_buff *skb);
int  FlowThread(void *pvArg);
int  CreateDevNode(void);
void DestroyDevNode(void);

static inline u16 flow_eth_type_trans(struct sk_buff *skb,
                                      struct net_device *dev)
{
    struct ethhdr *eth;

    /* Lincoln Tunnel -SMO for 2.6.27 */
    /* -SMO copied from eth_type_trans in <kern src>/net/ethernet/eth.c */
	skb->dev = dev;
    skb_reset_mac_header(skb);
    
	skb_pull(skb, ETH_HLEN);
    eth = eth_hdr(skb);

    if (__constant_ntohs(eth->h_proto) >= 1536)
        return eth->h_proto;

    return __constant_htons(ETH_P_802_2);
}

#endif /* _FLOWS_H_ */

